/*
Author:
Sreekumar Thaithara Balan 

Name:
control_system.cc

Version:
0.7

Last modified:
30-03-2008

Purpose:
1. Implemets a Control System to find out ideal step sizes for
   Random walk Metropolis alogrithm

   a. Acceptance Search: Adjust step sizes to obtain target acceptance rate
   b. Sigma Search: Modify the step sizes by the estimated std deviation

Implements:
1. Control_system::acceptance_search()-> control_system.h
2. Control_system::sigma_search()-> control_system.h

Todo:
1. Modify the sigma_search to Harrio et all 2003 method

References:
1. Rosentahl
2. H. Harrio, E. Saksman and J. Tamminen,

License:
GNU General Public License

This file is part of Exofit.

Exofit is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

Exofit is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Exofit; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.

*/

#include<iostream>
#include<fstream>
#include<cmath>
#include"ctrl_sys.h"

// ACCEPTANCE SEARCH

int Acceptance_search::search(std::ofstream& burn_file,bool index)
{

  // define a return val
  int ret_val=1;// normal exit

  // contstants for the sigma search
  const int max_paras=st->get_no_state();

  // set the initial sigmas with a scale factor
  std::vector<double> sigma_max;
  for(int i=0;i<max_paras;++i)
    {
      double s=st->get_para_range(i);
      st->set_para_step_size(i,s*sc);
      sigma_max.push_back(s*0.32);
    }  
 
  //find the annealing coefficients
  std::vector<double> anneal(max_lvls,0);
  for(int jj=0;jj<max_lvls;++jj)
    anneal[jj]=max_annl+(1-max_annl)*jj/(max_lvls-1);


  // begin optimization
  long int count=0;
  for(int j=0;j<max_lvls;++j)
    {
      std::cout<<"*****************************************************\n";
      std::cout<<"     Annealing with coeffient=["<<anneal[j]<<"]"<<"\n";
      std::cout<<"*****************************************************\n";
      bool check=true;

      while(check)
	{

	  // do samp_siz updates and find acceptance rate
	  std::vector<double> accept(max_paras,0);
	  for(int k=0;k<samp_siz;++k)
	    {
	      Update seq_update;
	      for(int i=0;i<max_paras;++i)
		{
		  seq_update.assign(max_paras,i,true);
		  accept[i]+=st->update(&seq_update);
		}
	      if(count%thn==0)
		{
		  if(index==true)
		    st->write_extract(burn_file,count);
		  else 
		    st->write_extract(burn_file);
		}

	      ++count;
	    }
	  for(int ii=0;ii<max_paras;++ii)
	    std::cout<<accept[ii]<<"/"<<(tg_acc*anneal[j]*samp_siz)<<"\t";
	  std::cout<<"\n";
	  for(int ii=0;ii<max_paras;++ii)
	    std::cout<<st->get_para_step_size(ii)<<"\t";
	  std::cout<<"\n";

	  // modify the step according to current accepatnce rate
	  for(int ii=0;ii<max_paras;++ii)
	    {
	      double sig=st->get_para_step_size(ii);
	      double adapt=0.05*sig;
	      if(accept[ii]>=(tg_acc+0.2*tg_acc)*anneal[j]*samp_siz)
		{
		  if( (sig+adapt)<sigma_max[ii] )
		    st->set_para_step_size(ii,sig+adapt);
		  else
		    st->set_para_step_size(ii,sig);
		}
	      else if(accept[ii]<=(tg_acc-0.2*tg_acc)*anneal[j]*samp_siz)
		if( (sig-adapt)>0)
		  st->set_para_step_size(ii,sig-adapt);
		else
		  st->set_para_step_size(ii,sig);
	    }

	  // exti condition
	  for(int iii=0;iii<max_paras;++iii)
	    {
	      if(accept[iii]<=(tg_acc-0.2*tg_acc)*anneal[j]*samp_siz)
		{
		  check=true;
		  ret_val=0;
		  break;
		}
	      else
		check=false;
	    }
	}// while(check)

      // extreme exit condition      
      if(count>=max_iter)
	{
	  ret_val=1;
	  break;
	}

    }// end of annealing
  burn_file.close();
  return ret_val;
}

// SIGMA SEARCH

int Sigma_search::search(std::ofstream& burn_file,
			  bool index_b, 
			  std::ofstream& diag_file,
			  bool index_d)
{

  // define a return value
  int ret_val=0;// normal exit

  // contstants for the sigma search
  const int max_paras=st->get_no_state();
  //std::cout<<max_paras<<"\n";
  
  // set the initial sigmas with a scale factor
  std::vector<double> sigma_max;
  for(int i=0;i<max_paras;++i)
    {
      double s=st->get_para_range(i);
      st->set_para_step_size(i,s*sc);
      sigma_max.push_back(s*0.32);
    }

  // estimate ideal step size from current sample
  long count=0;
  if(max_wind%2!=0)
    {
      std::cout<<"Maximum widows="<<max_wind<<"Should be an even number!\n";
      std::cout<<"Adding 1 to current value\n";
      max_wind=max_wind+1;
    }
  const int max_check=10*max_wind*samp_siz;
  bool check=true;
  std::vector< std::vector<double> >mu_matrix;
  while(check)
    {
      std::vector<double> mu(max_paras,0);
      std::vector<double> sigma(max_paras,0);
      std::vector<double> x(max_paras,0);
      std::vector<double> x2(max_paras,0);

      // update the state samp_siz times and find the current mean and variance
      for(int i=0;i<samp_siz;++i)
	{
	  Update seq_update;	  
	  for(int j=0;j<max_paras;++j)
	    {
	      seq_update.assign(max_paras,j,true);
	      st->update(&seq_update);
	      double tmp_val=st->get_para_val(j);
	      x[j] += tmp_val;
	      x2[j]+= tmp_val*tmp_val;
	      mu[j]=x[j]/(i+1);
	      double diff=(x2[j]-(i+1)*mu[j]*mu[j])/double(i);
	      if(diff<0)
		diff=diff*(-1);
	      if(i>0)
		sigma[j]=sqrt(diff);//i-> N-1
	    }

	  //write the burn file
	  if(count % thn==0)
	    {
	      if(index_b==true)		
		st->write_extract(burn_file,count);
	      else
		st->write_extract(burn_file);
	    }
	  ++count;
	} 

      
      
      if(index_d==true)
	diag_file<<count<<"\t";

      for(int j=0;j<max_paras;++j)
	{
	  st->set_para_step_size(j,sigma[j]);
	  //std::cout<<st->get_para_step_size(j)<<"|e|";
	  diag_file<<mu[j]<<"  "<<sigma[j]<<"    ";
	}
      //std::cout<<"\n";
      diag_file<<"\n";

      // calculate the variance of last max_wind means of samples
      mu_matrix.push_back(mu);
      int mat_size=mu_matrix.size();
      if(mat_size>=max_wind)
	{
	  std::vector<double> mu_est(max_paras,0);
	  std::vector<double> sig_est(max_paras,0);

	  for(int k=0;k<max_paras;++k)
	    {
	      std::vector<double> xx(max_paras,0);
	      std::vector<double> xx2(max_paras,0);

	      for(int i=1;i<=max_wind;++i)
		{
		  double tmp_v=mu_matrix[mat_size-i][k];
		  xx[k]+=tmp_v;
		  xx2[k]+=tmp_v*tmp_v;
		  mu_est[k]=xx[k]/(double)i;
		  if(i>1)
		    sig_est[k]=sqrt((xx2[k]-i*mu_est[k]*mu_est[k])/double(i-1));
		}
	      //std::cout<<sig_est[k]<<"|v|";
	    }
	  //std::cout<<"\n=================================================\n";

	  // exit condition: variance of last max_wind means < current variance
	  for(int j=0;j<max_paras;++j)
	    {
	      if(sig_est[j]>sigma[j])
		{
		  check=true;
		  break;
		}
	      else
		check=false;
	    }
	}//if(mat_size

      // extreme extit condition
      if(count>=max_check)
	{
	  ret_val=1;
	  break;
	}
    }  
  burn_file.close();
  diag_file.close();
  
  return ret_val;  
}


int Mean_monitored_run::run(std::ofstream& out_fil, bool index)
{
  // define a return value 
  int ret_val=0;

  const int max_paras=st->get_no_state();
  bool check=true;
  long count=0;
  std::vector< std::vector<double> >mu_matrix;

  while(check)
    {
      std::vector<double> x(max_paras,0);
      std::vector<double> x2(max_paras,0);
      std::vector<double> mu(max_paras,0);
      std::vector<double> sigma(max_paras,0);

      for(int i=0;i<samp_siz;++i)
	{

	  Update seq_update;
	  
	  for(int j=0;j<max_paras;++j)
	    {
	      seq_update.assign(max_paras,j,true);
	      st->update(&seq_update);
	      double tmp_val=st->get_para_val(j);
	      x[j] += tmp_val;
	      x2[j]+= tmp_val*tmp_val;
	      mu[j]=x[j]/(i+1);
	      if(i>0)
		sigma[j]=sqrt((x2[j]-(i+1)*mu[j]*mu[j])/(double)i);//i-> N-1
	    }
	  if(count>=2*samp_siz && count % thn==0)
	    st->write_extract(out_fil);

	  ++count;
	}
      mu_matrix.push_back(mu);
      int mat_size=mu_matrix.size();

      std::vector<double> mu_est(max_paras,0);
      std::vector<double> sig_est(max_paras,0);

      if(mat_size>=max_wind)
	{
	  std::vector<double> mean_first(max_paras,0);
	  std::vector<double> mean_last(max_paras,0);
	  std::vector<double> ratio(max_paras,0);
	  for(int j=0;j<max_paras;++j)
	    {
	      int middle=max_wind/2;
	      for(int i=1;i<middle;++i)
		{
		  mean_first[j]+=mu_matrix[mat_size-i][j];
		  mean_last[j]+=mu_matrix[mat_size-middle-i][j];
		}
	      mean_first[j]=mean_first[j]/middle;
	      mean_last[j]=mean_last[j]/middle;
	      ratio[j]=(mean_first[j]-mean_last[j])/mean_first[j];
	      if(ratio[j]<0)
		ratio[j]=(-1)*ratio[j];

	      //std::cout<<ratio[j]<<"||";

	    }
	  //std::cout<<"\n";

	  // exit condition
	  for(int k=0;k<max_paras;++k)
	    {
	      if(ratio[k]>=0.01)
		{
		  check=true;
		  break;
		}
	      else
		{
		  //std::cout<<"ratio["<<k<<"]="<<ratio[k]<<"||";
		  check=false;
		}
	    }
	  //std::cout<<"\n";
	      
	}

      // extreme exit condition
      if(count>=max_iter)
	{
	  ret_val=1;
	  break;
	}
    }

  return ret_val;
}
