/*
Author:
Sreekumar Thaithara Balan 

Name:
density.h

Version:
0.7

Last modified:
30-03-2008

Purpose:
1. Defines the interface for the following prior densitites and likelihood
   a. Uniform Prior
   b. Jeffreys Prior
   c. Modified Jeffreys

Implements:

Todo:
1. Find a way to implement fancy priors as in Gregory 2006

References:
1. Gregory 2006

License:
GNU General Public License
This file is part of Exofit.

Exofit is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

Exofit is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Exofit; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.

*/


#ifndef DENSITY_H
#define DENSITY_H

#include<vector>
#include<iostream>
#include"data.h"


class Density{

 public:

  //functions for examining the state of Density
  double get_min() const;
  double get_val() const;
  double get_max() const;

  //function for modifying the state of Density
  void set_elements(double,double,double);
    
  virtual double probability()=0;

 protected:

  double min;
  double val;
  double max;
};

class Likelihood{

 public:  
  int get_n_para() const;
  double get_para(int i) const;
  void set_data_file(Data* dt);
  void set_para(double p);
  void clear_para();
  virtual double likelihood()=0;

 protected:
  std::vector<double> par;
  Data* input_data;
};


class Null_density: public Density{
 public:
  double probability();
};

class Uniform: public Density{

 public:  
  double probability(); 
};


class Jeffreys: public Density{

 public:
  double probability();
};

class Modified_jeffreys: public Density{

 public:
  Modified_jeffreys(double);
  void set_break(double);
  double probability();
 
 private:
  double BREAK;
};

class N_planet_Gaussian: public Likelihood{

 public:
  N_planet_Gaussian(Data* dt,int n_planets);
  double likelihood();
 private:
  int n_pla;

};

/*********************** INLINES FOR abstract DENSITY ************************/

inline void Density::set_elements(double s,double v,double b)
{
  min=s;
  val=v;
  max=b;
}
inline double Density::get_min() const
{
  return min;
}
inline double Density::get_val() const
{
  return val;
}
inline double Density::get_max() const
{
  return max;
}

/********************** INLINE FOR NULL_DENSITY **************************/
inline double Null_density::probability()
{
  return 0.0;
}

/********************** INLINES FOR MODIFIED JEFFREYS ********************/
inline Modified_jeffreys::Modified_jeffreys(double brk)
{
  BREAK=brk;
}
inline void Modified_jeffreys::set_break(double n)
{
  BREAK=n;
}


/************************ INLINES FOR LIKELIHOOD **********************/

inline void Likelihood::set_data_file(Data* dt)
{
    input_data=dt;
}
inline void Likelihood::set_para(double p)
{
  //std::cout<<"setting para in lik="<<p<<"\n";
  par.push_back(p);
}
inline int Likelihood::get_n_para() const
{
  return par.size();
}
inline double Likelihood::get_para(int i) const
{
  return par[i];
}
inline void Likelihood::clear_para()
{
  par.clear();
}
/******************* INLINES FOR GAUSSIAN ******************************/
inline N_planet_Gaussian::N_planet_Gaussian(Data* dt,int n_planets)
{
  input_data=dt;
  n_pla=n_planets;
}


#endif // DENSITY_H
