/*
Author:
Sreekumar Thaithara Balan

Name:
mcmc.cc

Version:
0.7

Last modified:
30-03-2008

Purpose:
1. Definition of all the parameters,states,bonds,and control system
2. Define data files

Implements:
1. Mcmc::run()->mcmc.h
2. Parallel_tempering::initialise()->mcmc.h
3. Parallel_tempering::update()->mcmc.h

Todo:
1. Sequential update in parallel tempering

References:
1. Gregory Liu...
2. T. L. Graves

License:
GNU General Public License
This file is part of Exofit.

Exofit is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

Exofit is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Exofit; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.

*/


#include<iostream>
#include<fstream>
#include<cmath>
#include"mcmc.h"
#include"ctrl_sys.h"

void rv_planet_search(std::string rv_data_file,int n_planets)
{

  std::ifstream infile_data(rv_data_file.c_str(),std::ios::in);
  std::cout<<"\n\n                   Exofit version 2.0\n\n";

  if(!infile_data.is_open())
    {
      std::cout<<"Unable to open the data file: "<<rv_data_file<<"\n";
    }
  else
    {
      infile_data.close();
      // define the parameters here
      Parameter sys_velocity("V",-5000.0,0.0,5000.0,1000.0);
      Parameter period_1("T1",0.2,7500.0,15000.0,1500.0);
      Parameter amplitude_1("K1",0.000001,1000.0,2000.0,200.0);
      Parameter eccentricity_1("e1",0.0,0.5,0.99999,0.1);
      Parameter long_periastron_1("w1",0.0,3.14159,2*3.14159,0.628);
      Parameter periastron_pass_1("X1",0.0,0.5,0.9999,0.1);
      Parameter period_2("T2",0.2,7500.0,15000.0,1500.0);
      Parameter amplitude_2("K2",0.000001,1000.0,2000.0,200.0);
      Parameter eccentricity_2("e2",0.0,0.5,0.99999,0.1);
      Parameter long_periastron_2("w2",0.0,3.14159,2*3.14159,0.628);
      Parameter periastron_pass_2("X2",0.0,0.5,0.9999,0.1);
      Parameter noise_factor("s",0.0,1000.0,2000.0,200.0);

      //define the prior densities
      Uniform uniform;
      Jeffreys jeffreys;
      Modified_jeffreys mod_jeff(1.0);
      //Modified_jeffreys mod_jeff_low(0.00005);

      // Define the radial velocity data and read that file
      Data rv_data(rv_data_file,3);
      rv_data.read_file();

      std::cout<<"\nNumber of RV entries="<<rv_data.get_n_entries()<<"\n";
  
      // now make the data set start from zero  
      double start_val=rv_data.get_data_entry(0,0);
      for(int i=0;i<rv_data.get_n_entries();++i)
	{
	  double offset=rv_data.get_data_entry(i,0)-start_val;
	  rv_data.set_data_entry(i,0,offset);
	}

      //define the likelihood for 1 planet search
      N_planet_Gaussian Np_gauss_1(&rv_data,1);

      //define an mcmc state for 1 planet search
      State mcstate_1p;
      //add bonds
      mcstate_1p.add_bond(sys_velocity,&uniform);
      mcstate_1p.add_bond(period_1,&jeffreys);
      mcstate_1p.add_bond(amplitude_1,&mod_jeff);
      mcstate_1p.add_bond(eccentricity_1,&uniform);
      mcstate_1p.add_bond(long_periastron_1,&uniform);
      mcstate_1p.add_bond(periastron_pass_1,&uniform);
      mcstate_1p.add_bond(noise_factor,&mod_jeff);
      //add likelihood
      mcstate_1p.add_likelihood(&Np_gauss_1);

      //define likelihood for 2-planet search
      N_planet_Gaussian Np_gauss_2(&rv_data,2);
    
      //mcmc state for 2-planet serach
      State mcstate_2p;
      //add bonds
      mcstate_2p.add_bond(sys_velocity,&uniform);
      mcstate_2p.add_bond(period_1,&jeffreys);
      mcstate_2p.add_bond(amplitude_1,&mod_jeff);
      mcstate_2p.add_bond(eccentricity_1,&uniform);
      mcstate_2p.add_bond(long_periastron_1,&uniform);
      mcstate_2p.add_bond(periastron_pass_1,&uniform);
      mcstate_2p.add_bond(period_2,&jeffreys);
      mcstate_2p.add_bond(amplitude_2,&mod_jeff);
      mcstate_2p.add_bond(eccentricity_2,&uniform);
      mcstate_2p.add_bond(long_periastron_2,&uniform);
      mcstate_2p.add_bond(periastron_pass_2,&uniform);
      mcstate_2p.add_bond(noise_factor,&mod_jeff);
      //add likelihood
      mcstate_2p.add_likelihood(&Np_gauss_2);

      //read data from state data file "state.dat"
      std::string state_file="state.dat";
      std::ifstream infile_state(state_file.c_str(),std::ios::in);

      if(n_planets==1)
	{
	  if(infile_state.is_open())
	    {
	      std::cout<<"\nTaking parameters from \"state.dat\" \n\n";
	      bool succ=mcstate_1p.read_state_data(state_file);	 
	      if(!succ)
		{ 
		  std::cout<<"Using the default values for the state data\n";
		}
	    }
	  Mcmc mymc_1p(&mcstate_1p);
	  mymc_1p.run();
	}
      else if(n_planets==2)
	{
	  if(infile_state.is_open())
	    {
	      std::cout<<"\nTaking parameters from \"state.dat\" \n\n";
	      bool succ=mcstate_2p.read_state_data(state_file);	
	      if(!succ)
		{ 
		  std::cout<<"Using the default values for the state data\n";
		}  
	    }
	  Mcmc mymc_2p(&mcstate_2p);
	  mymc_2p.run();
	}
      else 
	std::cout<<"n_planets out of range[1,2]\n";
    }
}

//--------------------------mcmc starts here--------------------//
void Mcmc::run()
{

  // random numbers for state
  const gsl_rng_type* Ty;
  gsl_rng* rm;
  gsl_rng_env_setup();
  Ty=gsl_rng_default;
  rm=gsl_rng_alloc(Ty);
  unsigned long int sd=gsl_rng_uniform_int(rm,2500);
  mcmc_state->set_rands(gsl_rng_default,sd);

   //open the burn file
  std::ofstream brn_file;
  brn_file.open("burn.dat",std::ios::trunc);
  brn_file.close();
  brn_file.open("burn.dat",std::ios::app);

  //open the diagnostics file
  std::ofstream dg_file;
  dg_file.open("diag.dat",std::ios::trunc);
  dg_file.close();
  dg_file.open("diag.dat",std::ios::app);

  std::cout<<"\nSearching for ideal step sizes\n";
 
  Sigma_search sig_ser(mcmc_state,
		       0.12,//sclae
		       1000000,//max iterations
		       10000,// sample size
		       100,// thin
		       30);//windows
  int ser_succ=sig_ser.search(brn_file,false,dg_file,true);

  if(ser_succ==1)
    std::cout<<"CUATION ! Unsuccessful\n";
  else
    std::cout<<"\nSuccessful\n";

  brn_file.close();
  dg_file.close();
  
  //  define the output file
  std::ofstream out_file;
  out_file.open("extract.dat",std::ios::trunc); // clear data from outfile
  out_file.close();
  out_file.open("extract.dat",std::ios::app);// open in append mode
  mcmc_state->write_labels(out_file,false);	

  std::cout<<"\nStarting the main iteraions:\n";

  Mean_monitored_run m_run(mcmc_state,
			   1000000,// max iterations
			   10000,//sample size
			   100,// thin
			   50);// windows
  int m_succ=m_run.run(out_file,false);

  /*
  if(m_succ==1)
    {
      std::cout<<"\nCUATION ! One or more parameters shows divergence\n";
      std::cout<<"You may start the program with new random seed\n";
    }
  else
    std::cout<<"\nAll the parameters seem to be converged\n";
  */
  
  out_file.close();
  
}// end of mcmc





